---
title: "Tables de hachage"
date: "2022-01-12"
patat:
  eval:
    tai:
      command: fish
      fragment: false
      replace: true
    ccc:
      command: fish
      fragment: false
      replace: true
  images:
    backend: auto
...

# Tableau vs Table

## Tableau

* Chaque élément (ou valeur) est lié à un indice (la case du tableau).

```C 
annuaire tab[2] = {
    "+41 22 123 45 67", "+41 22 234 56 78", ...
};
tab[1] == "+41 22 123 45 67";
```

## Table

* Chaque élément (ou valeur) est lié à une clé.

```C 
annuaire tab = {
//  Clé   ,    Valeur
    "Paul",    "+41 22 123 45 67",
    "Orestis", "+41 22 234 56 78",
};
tab["Paul"]    == "+41 22 123 45 67";
tab["Orestis"] == "+41 22 234 56 78";
```

# Table

## Définition

Structure de données abstraite où chaque *valeur* (ou élément) est associée à une *clé* (ou
argument).

On parle de paires *clé-valeur* (*key-value pairs*).

## Donnez des exemples de telles paires

. . .

* Annuaire (nom-téléphone),
* Catalogue (objet-prix),
* Table de valeur fonctions (nombre-nombre),
* Index (nombre-page)
* ...

# Table

## Opérations principales sur les tables

* Insertion d'élément (`insert(clé, valeur)`{.C}), insère la paire `clé-valeur`
* Consultation (`get(clé)`{.C}), retourne la `valeur` correspondant à `clé`
* Suppression (`remove(clé)`{.C}), supprime la paire `clé-valeur`

## Structure de données / implémentation

Efficacité dépend de différents paramètres:

* taille (nombre de clé-valeurs maximal),
* fréquence d'utilisation (insertion, consultation, suppression),
* données triées/non-triées,
* ...

# Consultation séquentielle (`sequential_get`)

## Séquentielle

* table représentée par un (petit) tableau ou liste chaînée,
* types: `key_t` et `value_t` quelconques, et `key_value_t`

    ```C
    typedef struct {
        key_t key;
        value_t value;
    } key_value_t;
    ```
* on recherche l'existence de la clé séquentiellement dans le tableau, on
  retourne la valeur.

# Consultation séquentielle (`sequential_get`)

## Implémentation? Une idée?

. . .

```C
bool sequential_get(int n, key_value_t table[n], key_t key, 
    value_t *value) 
{
    int pos = n - 1;
    while (pos >= 0) {
        if (key ==  table[pos].key) {
            *value = table[pos].value;
            return true;
        }
        pos--;
    }
    return false;
}
```

. . .

## Inconvénient?

# Consultation séquentielle (`sequential_get`)

## Exercice: implémenter la même fonction avec une liste chaînée

Poster le résultat sur matrix.

# Consultation dichotomique (`binary_get`)

## Dichotomique

* table représentée par un (petit) tableau trié par les clés,
* types: `key_t` et `value_t` quelconques, et `key_value_t`
* on recherche l'existence de la clé par dichotomie dans le tableau, on
  retourne la valeur,
* les clés possèdent la notion d'ordre (`<, >, =` sont définis).

# Consultation dichotomique (`binary_get`)

\footnotesize

## Implémentation? Une idée?

. . .

```C
bool binary_get1(int n, value_key_t table[n], key_t key, value_t *value) {
    int top = n - 1, bottom = 0;
    while (top > bottom) { 
        int middle = (top + bottom) / 2;
        if (x > table[middle]) {
            bottom  = middle+1;
        } else {
            top = middle;
        }
    }
    if (key == table[top].key) {
        *value = table[top].value;
        return true;
    } else {
        return false;
    }
} 
```

# Consultation dichotomique (`binary_get`)

\footnotesize

## Autre implémentation

```C
bool binary_get2(int n, key_value_t table[n], key_t key, value_t *value) {
    int top = n - 1, bottom = 0;
    while (true) { 
        int middle = (top + bottom) / 2;
        if (key > table[middle].key) {
            bottom  = middle + 1;
        } else if (key < table[middle].key) {
            top = middle;
        } else {
            *value = table[middle].value;
            return true;
        }
        if (top < bottom) {
             break;
        }
    }
    return false;
}
```

## Quelle est la différence avec le code précédent?

# Transformation de clé (hashing)

## Problématique: Numéro AVS (13 chiffres)

* Format: 106.3123.8492.13

    ```
    Numéro AVS    | Nom
    0000000000000 | -------
    ...           | ...
    1063123849213 | Paul
    ...           | ...
    3066713878328 | Orestis
    ...           | ...
    9999999999999 | -------
    ```

## Quelle est la clé? Quelle est la valeur?

. . .

* Clé: Numéro AVS, Valeur: Nom.

## Nombre de clés? Nombre de citoyens? Rapport?

. . .

* $10^{13}$ clés, $10^8$ citoyens, $10^{-5}$ ($10^{-3}\%$ de la table est
  occupée) $\Rightarrow$ *inefficace*.
* Pire: $10^{13}$ entrées ne rentre pas dans la mémoire d'un
  ordinateur.

# Transformation de clé (hashing)

## Problématique 2: Identificateurs d'un programme

* Format: 8 caractères (simplification)

    ```
    Identificateur | Adresse
    aaaaaaaa       | -------
    ...            | ...
    resultat       | 3aeff
    compteur       | 4fedc
    ...            | ...
    zzzzzzzz       | -------
    ```

## Quelle est la clé? Quelle est la valeur?

. . .

* Clé: Identificateur, Valeur: Adresse.

## Nombre de clés? Nombre d'identificateur d'un programme? Rapport?

. . .

* $26^{8}\sim 2\cdot 10^{11}$ clés, $2000$ identificateurs, $10^{-8}$ ($10^{-6}\%$ de la table est
  occupée) $\Rightarrow$ *un peu inefficace*.

# Fonctions de transformation de clé (hash functions)

* La table est représentée avec un tableau.
* La taille du tableau est beaucoup plus petit que le nombre de clés.
* On produit un indice du tableau à partir d'une clé:
$$
h(key) = n,\quad n\in\mathbb{N}.
$$
En français: on transforme `key` en nombre entier qui sera l'indice dans le
tableau correspondant à `key`.

## La fonction de hash

* La taille du domaine des clés est beaucoup plus grand que le domaine des
  indices.
* Plusieurs indices peuvent correspondre à la **même clé**:
    * Il faut traiter les **collisions**.
* L'ensemble des indices doit être plus petit ou égal à la taille de la table.

## Une bonne fonction de hash

* Distribue uniformément les clés sur l'ensemble des indices.

# Fonctions de transformation de clés: exemples

## Méthode par troncature

\begin{align*}
&h: [0,9999]\rightarrow [0,9]\\
&h(key)=\mbox{troisième chiffre du nombre.}
\end{align*}

```
Key  | Index
0003 | 0
1123 | 2 \
1234 | 3  |-> collision.
1224 | 2 / 
1264 | 6 
```

## Quelle est la taille de la table?

. . .

C'est bien dix oui.

# Fonctions de transformation de clés: exemples

## Méthode par découpage

Taille de l'index: 3 chiffres.

```
key = 321 991 24 ->  321
                     991
                    + 24
                    ----
                    1336 -> index = 336
```

## Devinez l'algorithme?

. . .

On part de la gauche:

1. On découpe la clé en tranche de longueur égale à celle de l'index.
2. On somme les nombres obtenus.
3. On tronque à la longueur de l'index.

# Fonctions de transformation de clés: exemples

## Méthode multiplicative

Taille de l'index: 2 chiffres.

```
key = 5486 -> key^2 = 30096196 -> index = 96
```

On prend le carré de la clé et on garde les chiffres du milieu du résultat.

# Fonctions de transformation de clés: exemples

## Méthode par division modulo

Taille de l'index: `N` chiffres.

```
h(key) = key % N.
```

## Quelle doit être la taille de la table?

. . .

Oui comme vous le pensiez au moins `N`.

# Traitement des collisions

## La collision

```
key1 != key2, h(key1) == h(key2)
```

## Traitement (une idée?)

. . .

* La première clé occupe la place prévue dans le tableau.
* La deuxième (troisième, etc.) est placée ailleurs de façon **déterministe**.

Dans ce qui suit la taille de la table est `table_size`.

# La méthode séquentielle

* Quand l'index est déjà occupé on regarde sur la position suivante, jusqu'à en
  trouver une libre.

```C
index = h(key);
while (table[index].state == OCCUPIED && table[index].key != key) {
   index = (index + 1) % table_size; // attention à pas dépasser
}
table[index].key = key;
table[index].state = OCCUPIED;
```

## Problème?

. . .

* Regroupement d'éléments (clustering).

# Méthode linéaire

* Comme la méthode séquentielle mais on "saute" de `k`.

```C
index = h(key);
while (table[index].state == OCCUPIED && table[index].key != key) {
   index = (index + k) % table_size; // attention à pas dépasser
}
table[index].key = key;
table[index].state = OCCUPIED;
```

## Quelle valeur de `k` éviter?

. . .

* Une valeur où  `table_size` est multiple de `k`.

Cette méthode répartit mieux les regroupements au travers de la table.

# Méthode du double hashing

* Comme la méthode linéaire, mais `k = h2(key)` (variable).

```C
index = h(key);
while (table[index].state == OCCUPIED && table[index].key != key) {
   index = (index + h2(k)) % table_size; // attention à pas dépasser
}
table[index].key = key;
table[index].state = OCCUPIED;
```

## Quelle propriété doit avoir `h2`?

## Exemple

```C
h2(key) = (table_size - 2) - key % (table_size -2)
```

# Méthode pseudo-aléatoire

* Comme la méthode linéaire mais on génère `k` pseudo-aléatoirement.

```C
index = h(key);
while (table[index].state == OCCUPIED && table[index].key != key) {
   index = (index + random_number) % table_size; // attention à pas dépasser
}
table[index].key = key;
table[index].state = OCCUPIED;
```

## Comment s'assurer qu'on va bien retrouver la bonne clé?

. . .

* Le germe (seed) de la séquence pseudo-aléatoire doit être le même.
* Le germe à choisir est l'index retourné par `h(key)`.

    ```C
    srand(h(key));
    while {
        random_number = rand();
    }
    ```

# Méthode quadratique



# Méthode de chaînage
